from typing import TYPE_CHECKING, Optional, Pattern

if TYPE_CHECKING:

    from starlite.types import Scope, Scopes


def should_bypass_middleware(
    *,
    scope: "Scope",
    scopes: "Scopes",
    exclude_opt_key: Optional[str] = None,
    exclude_path_pattern: Optional[Pattern] = None,
) -> bool:
    """Determine weather a middleware should be bypassed.

    Args:
        scope: The ASGI scope.
        scopes: A set with the ASGI scope types that are supported by the middleware.
        exclude_opt_key: Key in `opt` with which a route handler can "opt-out" of a middleware.
        exclude_path_pattern: If this pattern matches scope["path"], the middleware should
            be bypassed.

    Returns:
        A boolean indicating if a middleware should be bypassed
    """
    if scope["type"] not in scopes:
        return True
    if exclude_opt_key and scope["route_handler"].opt.get(exclude_opt_key):
        return True
    if exclude_path_pattern and exclude_path_pattern.findall(scope["path"]):
        return True
    return False



from typing import TYPE_CHECKING, Any, Optional, cast

from anyio import Lock

from starlite.types import Empty
from starlite.utils.sync import AsyncCallable, is_async_callable

if TYPE_CHECKING:
    from typing import Type

    from starlite.signature import SignatureModel
    from starlite.types import AnyCallable


class Provide:
    __slots__ = ("dependency", "use_cache", "cache_per_request", "cache_key", "lock", "value", "signature_model")

    def __init__(
        self,
        dependency: "AnyCallable",
        use_cache: bool = False,
        cache_per_request: bool = False,
        cache_key: Optional[str] = None,
        sync_to_thread: bool = False,
    ) -> None:
        """A wrapper class used for dependency injection.

        Args:
            dependency: Callable to inject, can be a function, method or class.
            use_cache: Cache the dependency return value. Defaults to False.
            cache_per_request: Cache the dependency return value per request. Defaults to False.
            cache_key: Override the key for per request caching. Defaults to the function name.
            sync_to_thread: Run sync code in an async thread. Defaults to False.
        """
        self.dependency = cast("AnyCallable", AsyncCallable(dependency) if sync_to_thread else dependency)
        self.use_cache = use_cache
        self.cache_per_request = cache_per_request
        self.cache_key = cache_key if cache_key else getattr(dependency, "__name__", "anonymous")
        self.lock = Lock()
        self.value: Any = Empty
        self.signature_model: Optional["Type[SignatureModel]"] = None

    async def __call__(self, **kwargs: Any) -> Any:
        """Proxies call to 'self.proxy'."""

        if self.use_cache and self.value is not Empty:
            return self.value

        if is_async_callable(self.dependency):
            value = await self.dependency(**kwargs)
        else:
            value = self.dependency(**kwargs)

        if self.use_cache:
            self.value = value

        return value

    def __eq__(self, other: Any) -> bool:
        # check if memory address is identical, otherwise compare attributes
        return other is self or (
            isinstance(other, self.__class__)
            and other.dependency == self.dependency
            and other.use_cache == self.use_cache
            and other.value == self.value
        )



from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Set, Tuple, Type, Union, cast

from starlite.asgi.routing_trie.types import (
    ASGIHandlerTuple,
    PathParameterSentinel,
    create_node,
)
from starlite.asgi.utils import wrap_in_exception_handler
from starlite.types.internal_types import PathParameterDefinition

if TYPE_CHECKING:
    from starlite.app import Starlite
    from starlite.asgi.routing_trie.types import RouteTrieNode
    from starlite.routes import ASGIRoute, HTTPRoute, WebSocketRoute
    from starlite.types import ASGIApp, RouteHandlerType


def add_map_route_to_trie(
    app: "Starlite",
    root_node: "RouteTrieNode",
    route: Union["HTTPRoute", "WebSocketRoute", "ASGIRoute"],
    plain_routes: Set[str],
) -> "RouteTrieNode":
    """Adds a new route path (e.g. '/foo/bar/{param:int}') into the route_map
    tree.

    Inserts non-parameter paths ('plain routes') off the tree's root
    node. For paths containing parameters, splits the path on '/' and
    nests each path segment under the previous segment's node (see
    prefix tree / trie).

    Args:
        app: The Starlite app instance.
        root_node: The root trie node.
        route: The route that is being added.
        plain_routes: The set of plain routes.

    Returns:
        A RouteTrieNode instance.
    """
    current_node = root_node
    path = route.path

    is_mount = hasattr(route, "route_handler") and getattr(route.route_handler, "is_mount", False)  # type: ignore[union-attr]

    if not (route.path_parameters or is_mount):
        plain_routes.add(path)
        if path not in root_node["children"]:
            current_node["children"][path] = create_node()
        current_node = root_node["children"][path]
    else:
        for component in route.path_components:
            if isinstance(component, PathParameterDefinition):
                if component.type is Path:
                    current_node["is_path_type"] = True
                    break

                next_node_key: Union[Type[PathParameterSentinel], str] = PathParameterSentinel

            else:
                next_node_key = component

            if next_node_key not in current_node["children"]:
                current_node["children"][next_node_key] = create_node()

            current_node["child_keys"] = set(current_node["children"].keys())

            current_node = current_node["children"][next_node_key]

    configure_node(route=route, app=app, node=current_node)
    return current_node


def configure_node(
    app: "Starlite",
    route: Union["HTTPRoute", "WebSocketRoute", "ASGIRoute"],
    node: "RouteTrieNode",
) -> None:
    """Set required attributes and route handlers on route_map tree node.

    Args:
        app: The Starlite app instance.
        route: The route that is being added.
        node: The trie node being configured.

    Returns:
        None
    """
    from starlite.routes import HTTPRoute, WebSocketRoute

    if not node["path_parameters"]:
        node["path_parameters"] = route.path_parameters

    if isinstance(route, HTTPRoute):
        for method, handler_mapping in route.route_handler_map.items():
            handler, _ = handler_mapping
            node["asgi_handlers"][method] = ASGIHandlerTuple(
                asgi_app=build_route_middleware_stack(app=app, route=route, route_handler=handler),
                handler=handler,
            )

    elif isinstance(route, WebSocketRoute):
        node["asgi_handlers"]["websocket"] = ASGIHandlerTuple(
            asgi_app=build_route_middleware_stack(app=app, route=route, route_handler=route.route_handler),
            handler=route.route_handler,
        )

    else:
        node["asgi_handlers"]["asgi"] = ASGIHandlerTuple(
            asgi_app=build_route_middleware_stack(app=app, route=route, route_handler=route.route_handler),
            handler=route.route_handler,
        )
        node["is_asgi"] = True
        node["is_mount"] = route.route_handler.is_mount
        node["is_static"] = route.route_handler.is_static


def build_route_middleware_stack(
    app: "Starlite",
    route: Union["HTTPRoute", "WebSocketRoute", "ASGIRoute"],
    route_handler: "RouteHandlerType",
) -> "ASGIApp":
    """Constructs a middleware stack that serves as the point of entry for each
    route.

    Args:
        app: The Starlite app instance.
        route: The route that is being added.
        route_handler: The route handler that is being wrapped.

    Returns:
        An ASGIApp that is composed of a "stack" of middlewares.
    """
    from starlite.middleware.csrf import CSRFMiddleware

    # we wrap the route.handle method in the ExceptionHandlerMiddleware
    asgi_handler = wrap_in_exception_handler(
        debug=app.debug, app=route.handle, exception_handlers=route_handler.resolve_exception_handlers()  # type: ignore[arg-type]
    )

    if app.csrf_config:
        asgi_handler = CSRFMiddleware(app=asgi_handler, config=app.csrf_config)

    for middleware in route_handler.resolve_middleware():
        if hasattr(middleware, "__iter__"):
            handler, kwargs = cast("Tuple[Any, Dict[str, Any]]", middleware)
            asgi_handler = handler(app=asgi_handler, **kwargs)
        else:
            asgi_handler = middleware(app=asgi_handler)  # type: ignore

    # we wrap the entire stack again in ExceptionHandlerMiddleware
    return wrap_in_exception_handler(
        debug=app.debug,
        app=cast("ASGIApp", asgi_handler),
        exception_handlers=route_handler.resolve_exception_handlers(),
    )  # pyright: ignore



from typing import TYPE_CHECKING

from starlite.asgi.routing_trie.types import PathParameterSentinel
from starlite.exceptions import ImproperlyConfiguredException

if TYPE_CHECKING:
    from starlite.asgi.routing_trie.types import RouteTrieNode


def validate_node(node: "RouteTrieNode") -> None:
    """Recursively traverses the trie from the given node upwards.

    Args:
        node: A trie node.

    Raises:
        ImproperlyConfiguredException

    Returns:
        None
    """
    if node["is_asgi"] and bool(set(node["asgi_handlers"]).difference({"asgi"})):
        raise ImproperlyConfiguredException("ASGI handlers must have a unique path not shared by other route handlers.")

    if node["is_static"] and PathParameterSentinel in node["child_keys"]:
        raise ImproperlyConfiguredException("Path parameters cannot be configured for a static path.")

    for child in node["children"].values():
        validate_node(node=child)



import re
import secrets
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Pattern,
    Type,
    TypeVar,
    Union,
    cast,
)

from orjson import OPT_SERIALIZE_NUMPY, dumps, loads
from pydantic import BaseConfig, BaseModel, PrivateAttr, conint, conlist, constr
from typing_extensions import Literal

from starlite import ASGIConnection, Cookie, DefineMiddleware
from starlite.datastructures import MutableScopeHeaders
from starlite.middleware.base import MiddlewareProtocol
from starlite.middleware.utils import should_bypass_middleware
from starlite.types import Empty
from starlite.utils import default_serializer, get_serializer_from_scope

if TYPE_CHECKING:
    from starlite.types import ASGIApp, Message, Receive, Scope, ScopeSession, Send


ONE_DAY_IN_SECONDS = 60 * 60 * 24

ConfigT = TypeVar("ConfigT", bound="BaseBackendConfig")
ServerConfigT = TypeVar("ServerConfigT", bound="ServerSideSessionConfig")
BaseSessionBackendT = TypeVar("BaseSessionBackendT", bound="BaseSessionBackend")


class BaseBackendConfig(BaseModel):
    _backend_class: Type["BaseSessionBackend"] = PrivateAttr()

    """Configuration for Session middleware cookies."""

    class Config(BaseConfig):
        arbitrary_types_allowed = True

    key: constr(min_length=1, max_length=256) = "session"  # type: ignore[valid-type]
    """
    Key to use for the cookie inside the header,
    e.g. `session=<data>` where 'session' is the cookie key and <data> is the session data.

    Notes:
        - If a session cookie exceeds 4KB in size it is split. In this case the key will be of the format
            'session-{segment number}'.
    """
    max_age: conint(ge=1) = ONE_DAY_IN_SECONDS * 14  # type: ignore[valid-type]
    """Maximal age of the cookie before its invalidated."""
    scopes: conlist(Literal["http", "websocket"], min_items=1, max_items=2) = ["http", "websocket"]  # type: ignore[valid-type]
    """Scopes for the middleware - options are 'http' and 'websocket' with the default being both"""
    path: str = "/"
    """Path fragment that must exist in the request url for the cookie to be valid. Defaults to '/'."""
    domain: Optional[str] = None
    """Domain for which the cookie is valid."""
    secure: bool = False
    """Https is required for the cookie."""
    httponly: bool = True
    """Forbids javascript to access the cookie via 'Document.cookie'."""
    samesite: Literal["lax", "strict", "none"] = "lax"
    """Controls whether or not a cookie is sent with cross-site requests. Defaults to 'lax'."""
    exclude: Optional[Union[str, List[str]]] = None
    """A pattern or list of patterns to skip in the session middleware."""
    exclude_opt_key: str = "skip_session"
    """An identifier to use on routes to disable the session middleware for a particular route"""

    @property
    def middleware(self) -> DefineMiddleware:
        """Use this property to insert the config into a middleware list on one
        of the application layers.

        Examples:

            ```python
            from os import urandom

            from starlite import Starlite, Request, get
            from starlite.middleware.session import SessionCookieConfig

            session_config = SessionCookieConfig(secret=urandom(16))


            @get("/")
            def my_handler(request: Request) -> None:
                ...


            app = Starlite(route_handlers=[my_handler], middleware=[session_config.middleware])
            ```

        Returns:
            An instance of DefineMiddleware including 'self' as the config kwarg value.
        """
        return DefineMiddleware(SessionMiddleware, backend=self._backend_class(config=self))


class ServerSideSessionConfig(BaseBackendConfig):
    session_id_bytes: int = 32
    """
    Number of bytes used to generate a random session-ID
    """


class BaseSessionBackend(ABC, Generic[ConfigT]):
    __slots__ = ("config",)

    def __init__(self, config: ConfigT) -> None:
        """Abstract session backend defining the interface between a storage
        mechanism and the [SessionMiddleware][
        starlite.middleware.session.SessionMiddleware].

        This serves as the base class for all client- and server-side
        backends
        """
        self.config = config

    @staticmethod
    def serlialize_data(data: "ScopeSession", scope: Optional["Scope"] = None) -> bytes:
        """Serialize data into bytes for storage in the backend.

        Args:
            data: Session data of the current scope.
            scope: A scope, if applicable, from which to extract a serializer.

        Notes:
            - The serializer will be extracted from `scope` or fall back
                to [default_serializer][starlite.utils.default_serializer]

        Returns:
            `data` serialized as bytes.
        """
        serializer = (get_serializer_from_scope(scope) if scope else None) or default_serializer
        return dumps(data, default=serializer, option=OPT_SERIALIZE_NUMPY)

    @staticmethod
    def deserialize_data(data: Any) -> Dict[str, Any]:
        """Deserialize data into a dictionary for use in the application scope.

        Args:
            data: Data to be deserialized

        Returns:
            Deserialized data as a dictionary
        """
        return cast("Dict[str, Any]", loads(data))

    @abstractmethod
    async def store_in_message(
        self, scope_session: "ScopeSession", message: "Message", connection: ASGIConnection
    ) -> None:
        """Stores the necessary information in the outgoing `Message`

        Args:
            scope_session: Current session to store
            message: Outgoing send-message
            connection: Originating ASGIConnection containing the scope

        Returns:
            None
        """

    @abstractmethod
    async def load_from_connection(self, connection: ASGIConnection) -> Dict[str, Any]:
        """Load session data from a connection and return it as a dictionary to
        be used in the current application scope.

        Args:
            connection: An ASGIConnection instance

        Returns:
            The session data

        Notes:
            - This should not modify the connection's scope. The data returned by this
            method will be stored in the application scope by the middleware
        """


class ServerSideBackend(Generic[ServerConfigT], BaseSessionBackend[ServerConfigT]):
    __slots__ = ()

    def __init__(self, config: ServerConfigT) -> None:
        """Base class for server-side backends. Implements.

        [BaseSessionBackend][starlite.middleware.session.base.BaseSessionBackend] and defines and
        interface which subclasses can implement to facilitate the storage of session data
        """
        super().__init__(config=config)

    @abstractmethod
    async def get(self, session_id: str) -> Union[bytes, str, Dict[str, Any], None]:
        """Retrieve data associated with `session_id`.

        Args:
            session_id: The session-ID

        Returns:
            The session data, if existing, otherwise `None`.
        """

    @abstractmethod
    async def set(self, session_id: str, data: bytes) -> None:
        """Store `data` under the `session_id` for later retrieval.

        If there is already data associated with `session_id`, replace
        it with `data` and reset its expiry time

        Args:
            session_id: The session-ID
            data: Serialized session data

        Returns:
            None
        """

    @abstractmethod
    async def delete(self, session_id: str) -> None:
        """Delete the data associated with `session_id`. Fails silently if no
        such session-ID exists.

        Args:
            session_id: The session-ID

        Returns:
            None
        """

    @abstractmethod
    async def delete_all(self) -> None:
        """Delete all session data stored within this backend.

        Returns:
            None
        """

    def generate_session_id(self) -> str:
        """Generate a new session-ID, with n=[session_id_bytes][starlite.middle
        ware.session.base.ServerSideSessionConfig.session_id_bytes] random
        bytes.

        Returns:
            A session-ID
        """
        return secrets.token_hex(self.config.session_id_bytes)

    async def store_in_message(
        self, scope_session: "ScopeSession", message: "Message", connection: ASGIConnection
    ) -> None:
        """Stores the necessary information in the outgoing `Message` by
        setting a cookie containing the session-ID. If the session is empty, a
        null-cookie will be set. Otherwise, the serialised data will be stored
        using [set][starlite.middleware.session.base.ServerSideBackend.set],
        under the current session-id. If no session-ID exists, a new ID will be
        generated using [generate_session_id][starlite.middleware.session.base.
        ServerSideBackend.generate_session_id].

        Args:
            scope_session: Current session to store
            message: Outgoing send-message
            connection: Originating ASGIConnection containing the scope

        Returns:
            None
        """
        scope = connection.scope
        headers = MutableScopeHeaders.from_message(message)
        session_id = connection.cookies.get(self.config.key)
        if session_id == "null":
            session_id = None
        if not session_id:
            session_id = self.generate_session_id()

        cookie_params = self.config.dict(
            exclude_none=True,
            exclude={"secret", "key"} | set(self.config.__fields__) - set(BaseBackendConfig.__fields__),
        )

        if scope_session is Empty:
            await self.delete(session_id)
            headers.add(
                "Set-Cookie",
                Cookie(value="null", key=self.config.key, expires=0, **cookie_params).to_header(header=""),
            )
        else:
            serialised_data = self.serlialize_data(scope_session, scope)
            await self.set(session_id=session_id, data=serialised_data)

            headers["Set-Cookie"] = Cookie(value=session_id, key=self.config.key, **cookie_params).to_header(header="")

    async def load_from_connection(self, connection: ASGIConnection) -> Dict[str, Any]:
        """Load session data from a connection and return it as a dictionary to
        be used in the current application scope.

        The session-ID will be gathered from a cookie with the key set in the
        [configuration][starlite.middleware.session.base.BaseBackendConfig.key]. If
        a cookie is found, its value will be used as the session-ID and data associated
        with this ID will be loaded using [get][starlite.middleware.session.base.ServerSideBackend.get].
        If no cookie was found or no data was loaded from the store, this will return an
        empty dictionary.

        Args:
            connection: An ASGIConnection instance

        Returns:
            The current session data
        """
        session_id = connection.cookies.get(self.config.key)
        if session_id:
            data = await self.get(session_id)
            if isinstance(data, dict):
                return data
            if data is not None:
                return self.deserialize_data(data)
        return {}


class SessionMiddleware(MiddlewareProtocol, Generic[BaseSessionBackendT]):
    __slots__ = ("backend", "_exclude_pattern", "_exclude_opt_key")

    def __init__(self, app: "ASGIApp", backend: BaseSessionBackendT) -> None:
        """Starlite session middleware for storing session data.

        Args:
            app: An ASGI application
            backend: A [BaseSessionBackend][starlite.middleware.session.base.BaseSessionBackend]
                instance used to store and retrieve session data
        """

        self.app = app
        self.backend = backend
        self._exclude_pattern: Optional[Pattern[str]] = None
        self._exclude_opt_key = backend.config.exclude_opt_key

        if backend.config.exclude:
            exclude = backend.config.exclude
            self._exclude_pattern = re.compile("|".join(exclude)) if isinstance(exclude, list) else re.compile(exclude)

    def create_send_wrapper(self, connection: ASGIConnection) -> Callable[["Message"], Awaitable[None]]:
        """
        Creates a wrapper for the ASGI send function, which handles setting the cookies on the outgoing response.
        Args:
            connection: ASGIConnection

        Returns:
            None
        """

        async def wrapped_send(message: "Message") -> None:
            """A wrapper around the send function, declared in local scope to
            use closure values.

            Args:
                message: An ASGI message.

            Returns:
                None
            """
            if message["type"] != "http.response.start":
                await connection.send(message)
                return

            scope_session = connection.scope.get("session")

            await self.backend.store_in_message(scope_session, message, connection)
            await connection.send(message)

        return wrapped_send

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        if not should_bypass_middleware(
            scope=scope,
            scopes=self.backend.config.scopes,
            exclude_opt_key=self._exclude_opt_key,
            exclude_path_pattern=self._exclude_pattern,
        ):
            connection = ASGIConnection[Any, Any, Any](scope, receive=receive, send=send)
            scope["session"] = await self.backend.load_from_connection(connection)

            await self.app(scope, receive, self.create_send_wrapper(connection))
        else:
            await self.app(scope, receive, send)



from typing import TYPE_CHECKING, Any, Dict, Iterable, Union

from anyio import create_task_group
from pydantic.fields import FieldInfo

from starlite.constants import EXTRA_KEY_IS_DEPENDENCY, EXTRA_KEY_SKIP_VALIDATION

if TYPE_CHECKING:
    from typing_extensions import TypeGuard

    from starlite.connection import Request, WebSocket
    from starlite.kwargs import Dependency, KwargsModel


def is_dependency_field(val: Any) -> "TypeGuard[FieldInfo]":
    """Determine if a value is a `FieldInfo` instance created via the
    `Dependency()` function.

    Args:
        val(Any): value to be tested

    Returns:
        `True` if `val` is `FieldInfo` created by [`Dependency()`][starlite.params.Dependency] function.
    """
    return isinstance(val, FieldInfo) and bool(val.extra.get(EXTRA_KEY_IS_DEPENDENCY))


def should_skip_dependency_validation(val: Any) -> bool:
    """Determine if a value is a `FieldInfo` instance created via the
    `Dependency()` function set with ` skip_validation=True`.

    Args:
        val(Any): value to be tested

    Returns:
        `True` if `val` is `FieldInfo` created by [`Dependency()`][starlite.params.Dependency] function and
        `skip_validation=True` is set.
    """
    return is_dependency_field(val) and bool(val.extra.get(EXTRA_KEY_SKIP_VALIDATION))


async def _resolve_dependency_into_kwargs(
    model: "KwargsModel", dependency: "Dependency", connection: Union["WebSocket", "Request"], kwargs: Dict[str, Any]
) -> None:
    """Helper function to store the result of resolve_dependency in the
    kwargs."""
    kwargs[dependency.key] = await model.resolve_dependency(dependency=dependency, connection=connection, **kwargs)


async def resolve_dependencies_concurrently(
    model: "KwargsModel",
    dependencies: Iterable["Dependency"],
    connection: Union["WebSocket", "Request"],
    kwargs: Dict[str, Any],
) -> None:
    """Resolve the dependencies concurrently. The passed in kwargs will be
    modified with the results.

    Args:
        model: An instance of [KwargsModel][starlite.KwargsModel].
        dependencies: An iterable for instances of [Dependency][starlite.kwargs.Dependency].
        connection: An instance of [Request][starlite.connection.Request] or [WebSocket][starlite.connection.WebSocket].
        kwargs: Kwargs to pass to dependencies
    """
    async with create_task_group() as task_group:
        for dependency in dependencies:
            task_group.start_soon(_resolve_dependency_into_kwargs, model, dependency, connection, kwargs)



from typing import List, Optional, Type, Union

from pydantic import BaseModel, conint
from typing_extensions import Literal

from starlite.middleware.compression import CompressionMiddleware


class CompressionConfig(BaseModel):
    """Configuration for response compression.

    To enable response compression, pass an instance of this class to
    the [Starlite][starlite.app.Starlite] constructor using the
    'compression_config' key.
    """

    backend: Literal["gzip", "brotli"]
    """
        Literal of "gzip" or "brotli"
    """
    minimum_size: conint(gt=0) = 500  # type: ignore[valid-type]
    """
        Minimum response size (bytes) to enable compression, affects all backends.
    """
    gzip_compress_level: conint(ge=0, le=9) = 9  # type: ignore[valid-type]
    """
        Range [0-9], see [official docs](https://docs.python.org/3/library/gzip.html).
    """
    brotli_quality: conint(ge=0, le=11) = 5  # type: ignore[valid-type]
    """
        Range [0-11], Controls the compression-speed vs compression-density tradeoff. The higher the quality, the slower
        the compression.
    """
    brotli_mode: Literal["generic", "text", "font"] = "text"
    """
        MODE_GENERIC, MODE_TEXT (for UTF-8 format text input, default) or MODE_FONT (for WOFF 2.0).
    """
    brotli_lgwin: conint(ge=10, le=24) = 22  # type: ignore[valid-type]
    """
        Base 2 logarithm of size. Range is 10 to 24. Defaults to 22.
    """
    brotli_lgblock: Literal[0, 16, 17, 18, 19, 20, 21, 22, 23, 24] = 0
    """
        Base 2 logarithm of the maximum input block size. Range is 16 to 24. If set to 0, the value will be set based
        on the quality. Defaults to 0.
    """
    brotli_gzip_fallback: bool = True
    """
        Use GZIP if Brotli is not supported.
    """
    middleware_class: Type[CompressionMiddleware] = CompressionMiddleware
    """
    Middleware class to use, should be a subclass of CompressionMiddleware.
    """
    exclude: Optional[Union[str, List[str]]] = None
    """
    An identifier to use on routes to disable authentication for a particular route.
    """
    exclude_opt_key: Optional[str] = None
    """
    A pattern or list of patterns to skip in the authentication middleware.
    """



import re
from dataclasses import dataclass
from time import time
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Pattern,
    Tuple,
    Type,
    cast,
)

from orjson import dumps, loads
from pydantic import BaseModel, validator
from typing_extensions import Literal

from starlite.connection import Request
from starlite.datastructures import MutableScopeHeaders
from starlite.enums import ScopeType
from starlite.exceptions import TooManyRequestsException
from starlite.middleware.base import DefineMiddleware
from starlite.types import Message, SyncOrAsyncUnion
from starlite.utils import AsyncCallable

if TYPE_CHECKING:
    from typing import Awaitable

    from starlite.cache import Cache
    from starlite.types import ASGIApp, Receive, Scope, Send

DurationUnit = Literal["second", "minute", "hour", "day"]

DURATION_VALUES: Dict[DurationUnit, int] = {"second": 1, "minute": 60, "hour": 3600, "day": 86400}


@dataclass
class CacheObject:
    __slots__ = ("history", "reset")

    history: List[int]
    reset: int


class RateLimitMiddleware:
    __slots__ = (
        "app",
        "cache",
        "check_throttle_handler",
        "exclude",
        "max_requests",
        "unit",
        "request_quota",
        "config",
    )

    cache: "Cache"

    def __init__(self, app: "ASGIApp", config: "RateLimitConfig") -> None:
        """

        Args:
            app: The 'next' ASGI app to call.
            config: An instance of RateLimitConfig.
        """
        self.app = app
        self.check_throttle_handler = cast(
            "Optional[Callable[[Request], Awaitable[bool]]]", config.check_throttle_handler
        )
        self.exclude: Optional[Pattern[str]] = re.compile("|".join(config.exclude)) if config.exclude else None
        self.max_requests: int = config.rate_limit[1]
        self.config = config
        self.unit: DurationUnit = config.rate_limit[0]

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """
        Args:
            scope: The ASGI connection scope.
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None
        """
        if scope["type"] == ScopeType.HTTP:
            if not hasattr(self, "cache"):
                self.cache = scope["app"].cache

            request: "Request[Any, Any]" = scope["app"].request_class(scope)
            if await self.should_check_request(request=request):
                key = self.cache_key_from_request(request=request)
                cache_object = await self.retrieve_cached_history(key)
                if len(cache_object.history) >= self.max_requests:
                    raise TooManyRequestsException(
                        headers=self.create_response_headers(cache_object=cache_object)
                        if self.config.set_rate_limit_headers
                        else None
                    )
                await self.set_cached_history(key=key, cache_object=cache_object)
                if self.config.set_rate_limit_headers:
                    send = self.create_send_wrapper(send=send, cache_object=cache_object)

        await self.app(scope, receive, send)

    def create_send_wrapper(self, send: "Send", cache_object: CacheObject) -> "Send":
        """Creates a send function that wraps the original send to inject
        response headers.

        Args:
            send: The ASGI send function.
            cache_object: A CacheObject instance.

        Returns:
            Send wrapper callable.
        """

        async def send_wrapper(message: "Message") -> None:
            """

            Args:
                message: An ASGI 'Message'

            Returns:
                None
            """
            if message["type"] == "http.response.start":
                message.setdefault("headers", [])
                headers = MutableScopeHeaders(message)
                for key, value in self.create_response_headers(cache_object=cache_object).items():
                    headers.add(key, value)
            await send(message)

        return send_wrapper

    def cache_key_from_request(self, request: "Request[Any, Any]") -> str:
        """

        Args:
            request: A [Request][starlite.connection.Request] instance.

        Returns:
            A cache key.
        """
        host = request.client.host if request.client else "anonymous"
        identifier = request.headers.get("X-Forwarded-For") or request.headers.get("X-Real-IP") or host
        return f"{type(self).__name__}::{identifier}"

    async def retrieve_cached_history(self, key: str) -> CacheObject:
        """Retrieves a list of time stamps for the given duration unit.

        Args:
            key: Cache key.

        Returns:
            An instance of CacheObject.
        """
        duration = DURATION_VALUES[self.unit]
        now = int(time())
        cached_string = await self.cache.get(key)
        if cached_string:
            cache_object = CacheObject(**loads(cached_string))
            if cache_object.reset <= now:
                return CacheObject(history=[], reset=now + duration)

            while cache_object.history and cache_object.history[-1] <= now - duration:
                cache_object.history.pop()
            return cache_object

        return CacheObject(history=[], reset=now + duration)

    async def set_cached_history(self, key: str, cache_object: CacheObject) -> None:
        """Stores history extended with the current timestamp in cache.

        Args:
            key: Cache key.
            cache_object: An instance of CacheObject.

        Returns:
            None
        """
        cache_object.history = [int(time()), *cache_object.history]
        await self.cache.set(key, dumps(cache_object), expiration=DURATION_VALUES[self.unit])

    async def should_check_request(self, request: "Request[Any, Any]") -> bool:
        """

        Args:
            request: A [Request][starlite.connection.Request] instance.

        Returns:
            Boolean dictating whether the request should be checked for rate-limiting.
        """
        if not self.exclude or not self.exclude.findall(request.url.path):
            if self.check_throttle_handler:
                return await self.check_throttle_handler(request)
            return True
        return False

    def create_response_headers(self, cache_object: CacheObject) -> Dict[str, str]:
        """Creates ratelimit response headers.

        Notes:
            * see the [IETF RateLimit draft][https://datatracker.ietf.org/doc/draft-ietf-httpapi-ratelimit-headers/]

        Args:
            cache_object: An instance of Cache Object.

        Returns:
            A dict of http headers.
        """
        remaining_requests = str(
            len(cache_object.history) - self.max_requests if len(cache_object.history) <= self.max_requests else 0
        )

        return {
            self.config.rate_limit_policy_header_key: f"{self.max_requests}; w={DURATION_VALUES[self.unit]}",
            self.config.rate_limit_limit_header_key: str(self.max_requests),
            self.config.rate_limit_remaining_header_key: remaining_requests,
            self.config.rate_limit_reset_header_key: str(int(time()) - cache_object.reset),
        }


class RateLimitConfig(BaseModel):
    rate_limit: Tuple[DurationUnit, int]
    """A tuple containing a time unit (second, minute, hour, day) and quantity, e.g. ("day", 1) or ("minute", 5)."""
    exclude: Optional[List[str]] = None
    """List of patterns to skip in the authentication middleware."""
    check_throttle_handler: Optional[Callable[[Request[Any, Any]], SyncOrAsyncUnion[bool]]] = None
    """
    Handler callable that receives the request instance, returning a boolean dictating whether or not the
    request should be checked for rate limiting.
    """
    middleware_class: Type[RateLimitMiddleware] = RateLimitMiddleware
    """The middleware class to use."""
    set_rate_limit_headers: bool = True
    """Boolean dictating whether to set the rate limit headers on the response."""
    rate_limit_policy_header_key: str = "RateLimit-Policy"
    """Key to use for the rate limit policy header."""
    rate_limit_remaining_header_key: str = "RateLimit-Remaining"
    """Key to use for the rate limit remaining header."""
    rate_limit_reset_header_key: str = "RateLimit-Reset"
    """Key to use for the rate limit reset header."""
    rate_limit_limit_header_key: str = "RateLimit-Limit"
    """Key to use for the rate limit limit header."""
    cache_key_builder: Optional[Callable[[Request], str]] = None

    @validator("check_throttle_handler")
    def validate_check_throttle_handler(cls, value: Callable) -> Callable:  # pylint: disable=no-self-argument
        """

        Args:
            value: A callable.

        Returns:
            An instance of AsyncCallable
        """
        return AsyncCallable(value)

    @property
    def middleware(self) -> DefineMiddleware:
        """Use this property to insert the config into a middleware list on one
        of the application layers.

        Examples:

            ```python
            from starlite import Starlite, Request, get
            from starlite.middleware import RateLimitConfig

            # limit to 10 requests per minute, excluding the schema path
            throttle_config = RateLimitConfig(rate_limit=("minute", 10), exclude=["/schema"])


            @get("/")
            def my_handler(request: Request) -> None:
                ...


            app = Starlite(route_handlers=[my_handler], middleware=[throttle_config.middleware])
            ```

        Returns:
            An instance of DefineMiddleware including 'self' as the config kwarg value.
        """
        return DefineMiddleware(self.middleware_class, config=self)



from collections import deque
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from pathlib import Path
from re import sub
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Deque,
    Dict,
    List,
    Set,
    Tuple,
    Type,
    Union,
)
from uuid import UUID

from pydantic.datetime_parse import (
    parse_date,
    parse_datetime,
    parse_duration,
    parse_time,
)

from starlite.asgi.routing_trie.types import PathParameterSentinel
from starlite.enums import ScopeType
from starlite.exceptions import (
    MethodNotAllowedException,
    NotFoundException,
    ValidationException,
)
from starlite.utils import normalize_path

if TYPE_CHECKING:
    from starlite.asgi.routing_trie.types import ASGIHandlerTuple, RouteTrieNode
    from starlite.types import Scope
    from starlite.types.internal_types import PathParameterDefinition

parsers_map: Dict[Any, Callable[[Any], Any]] = {
    str: str,
    float: float,
    int: int,
    Decimal: Decimal,
    UUID: UUID,
    Path: lambda x: Path(sub("//+", "", (x.lstrip("/")))),
    date: parse_date,
    datetime: parse_datetime,
    time: parse_time,
    timedelta: parse_duration,
}


def traverse_route_map(
    current_node: "RouteTrieNode",
    path: str,
    path_components: Deque[Union[str, Type[PathParameterSentinel]]],
    path_params: List[str],
    scope: "Scope",
) -> Tuple["RouteTrieNode", List[str]]:
    """Traverses the application route mapping and retrieves the correct node
    for the request url.

    Args:
        current_node: A trie node.
        path: The request's path.
        path_components: A list of ordered path components.
        path_params: A list of extracted path parameters.
        scope: The ASGI connection scope.

    Raises:
         NotFoundException: if no correlating node is found.

    Returns:
        A tuple containing the target RouteMapNode and a list containing all path parameter values.
    """

    if current_node["is_mount"]:
        if current_node["is_static"] and not (path_components and path_components[0] in current_node["child_keys"]):
            # static paths require an ending slash.
            scope["path"] = normalize_path("/".join(path_components) + "/")  # type: ignore[arg-type]
            return current_node, path_params
        if not current_node["is_static"]:
            scope["path"] = normalize_path("/".join(path_components))  # type: ignore[arg-type]
            return current_node, path_params

    if current_node["is_path_type"]:
        path_params.append(normalize_path("/".join(path_components)))  # type: ignore[arg-type]
        return current_node, path_params

    has_path_param = PathParameterSentinel in current_node["child_keys"]

    if not path_components:
        if has_path_param or not current_node["asgi_handlers"]:
            raise NotFoundException()
        return current_node, path_params

    component = path_components.popleft()

    if component in current_node["child_keys"]:
        return traverse_route_map(
            current_node=current_node["children"][component],
            path=path,
            path_components=path_components,
            path_params=path_params,
            scope=scope,
        )

    if has_path_param:
        path_params.append(component)  # type: ignore[arg-type]

        return traverse_route_map(
            current_node=current_node["children"][PathParameterSentinel],
            path=path,
            path_components=path_components,
            path_params=path_params,
            scope=scope,
        )

    raise NotFoundException()


def parse_path_parameters(
    path_parameter_definitions: List["PathParameterDefinition"], request_path_parameter_values: List[str]
) -> Dict[str, Any]:
    """Parses path parameters into their expected types.

    Args:
        path_parameter_definitions: A list of [PathParameterDefinition][starlite.route.base.PathParameterDefinition] instances
        request_path_parameter_values: A list of raw strings sent as path parameters as part of the request

    Raises:
        ValidationException: if path parameter parsing fails

    Returns:
        A dictionary mapping path parameter names to parsed values
    """
    result: Dict[str, Any] = {}

    try:
        for idx, parameter_definition in enumerate(path_parameter_definitions):
            raw_param_value = request_path_parameter_values[idx]
            parser = parsers_map[parameter_definition.type]
            result[parameter_definition.name] = parser(raw_param_value)
        return result
    except (ValueError, TypeError, KeyError) as e:  # pragma: no cover
        raise ValidationException(f"unable to parse path parameters {','.join(request_path_parameter_values)}") from e


def parse_scope_to_route(root_node: "RouteTrieNode", scope: "Scope", plain_routes: Set[str]) -> "ASGIHandlerTuple":
    """Given a scope object, retrieve the asgi_handlers and is_mount boolean
    values from correct trie node.

    Args:
        root_node: The root trie node.
        scope: The ASGI scope instance.
        plain_routes: The set of plain routes.

    Raises:
        MethodNotAllowedException: if no matching method is found.

    Returns:
        A tuple containing the stack of middlewares and the route handler that is wrapped by it.
    """

    path = scope["path"].strip().rstrip("/") or "/"
    scope["path_params"] = {}

    if path in plain_routes:
        current_node: "RouteTrieNode" = root_node["children"][path]
    else:
        current_node, path_params = traverse_route_map(
            current_node=root_node,
            path=path,
            path_components=deque([component for component in path.split("/") if component]),
            path_params=[],
            scope=scope,
        )
        if path_params:
            scope["path_params"] = parse_path_parameters(
                path_parameter_definitions=current_node["path_parameters"],
                request_path_parameter_values=path_params,
            )
    try:
        if current_node["is_asgi"]:
            return current_node["asgi_handlers"]["asgi"]

        if scope["type"] == ScopeType.HTTP:
            return current_node["asgi_handlers"][scope["method"]]

        return current_node["asgi_handlers"]["websocket"]
    except KeyError as e:
        raise MethodNotAllowedException() from e



from dataclasses import dataclass
from enum import Enum
from typing import Any, Generic, Optional, TypeVar, Union, cast

T = TypeVar("T")


def get_name(value: Any) -> str:
    """Helper to get the '__name__' dunder of a value.

    Args:
        value: An arbitrary value.

    Returns:
        A name string.
    """

    if hasattr(value, "__name__"):
        return cast("str", value.__name__)
    return type(value).__name__


def get_enum_string_value(value: Union[Enum, str]) -> str:
    """A helper function to return the string value of a string enum.

    See: https://github.com/starlite-api/starlite/pull/633#issuecomment-1286519267

    Args:
        value: An enum or string.

    Returns:
        A string.
    """
    return cast("str", value.value) if isinstance(value, Enum) else value


@dataclass
class Ref(Generic[T]):
    """A helper class that encapsulates a value."""

    __slots__ = ("value",)

    value: Optional[T]
    """The value wrapped by the ref."""



from copy import copy
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Type,
    TypeVar,
    Union,
    cast,
)

from pydantic import BaseConfig, Extra, validate_arguments
from pydantic.fields import ModelField, Undefined

from starlite.constants import EXTRA_KEY_REQUIRED, EXTRA_KEY_VALUE_TYPE
from starlite.datastructures.provide import Provide
from starlite.exceptions import ImproperlyConfiguredException
from starlite.types import (
    Dependencies,
    Empty,
    EmptyType,
    ExceptionHandlersMap,
    Guard,
    Middleware,
    ParametersMap,
)
from starlite.utils import AsyncCallable, normalize_path

if TYPE_CHECKING:
    from starlite.connection import ASGIConnection
    from starlite.controller import Controller
    from starlite.router import Router
    from starlite.signature import SignatureModel
    from starlite.types import AnyCallable

T = TypeVar("T", bound="BaseRouteHandler")


class ParameterConfig(BaseConfig):
    extra = Extra.allow


class BaseRouteHandler(Generic[T]):
    __slots__ = (
        "_resolved_dependencies",
        "_resolved_guards",
        "_resolved_layered_parameters",
        "dependencies",
        "exception_handlers",
        "fn",
        "guards",
        "middleware",
        "name",
        "opt",
        "owner",
        "paths",
        "signature_model",
    )

    @validate_arguments(config={"arbitrary_types_allowed": True})
    def __init__(
        self,
        path: Union[Optional[str], Optional[List[str]]] = None,
        *,
        dependencies: Optional[Dependencies] = None,
        exception_handlers: Optional[ExceptionHandlersMap] = None,
        guards: Optional[List[Guard]] = None,
        middleware: Optional[List[Middleware]] = None,
        name: Optional[str] = None,
        opt: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        self._resolved_dependencies: Union[Dict[str, Provide], EmptyType] = Empty
        self._resolved_guards: Union[List[Guard], EmptyType] = Empty
        self._resolved_layered_parameters: Union[Dict[str, "ModelField"], EmptyType] = Empty
        self.dependencies = dependencies
        self.exception_handlers = exception_handlers
        self.fn: Optional["AnyCallable"] = None
        self.guards = guards
        self.middleware = middleware
        self.name = name
        self.opt: Dict[str, Any] = opt or {}
        self.owner: Optional[Union["Controller", "Router"]] = None
        self.signature_model: Optional[Type["SignatureModel"]] = None
        self.paths = (
            {normalize_path(p) for p in path}
            if path and isinstance(path, list)
            else {normalize_path(path or "/")}  # type: ignore
        )
        self.opt.update(**kwargs)

    @property
    def dependency_name_set(self) -> Set[str]:
        """The set of all dependency names provided in the handler's ownership
        layers."""
        layered_dependencies = (layer.dependencies or {} for layer in self.ownership_layers)
        return {name for layer in layered_dependencies for name in layer.keys()}

    @property
    def ownership_layers(self) -> List[Union[T, "Controller", "Router"]]:
        """Returns the handler layers from the app down to the route handler.

        app -> ... -> route handler
        """
        layers = []

        cur: Any = self
        while cur:
            layers.append(cur)
            cur = cur.owner

        return list(reversed(layers))

    def resolve_layered_parameters(self) -> Dict[str, "ModelField"]:
        """Returns all parameters declared above the handler, transforming them
        into pydantic ModelField instances."""
        if self._resolved_layered_parameters is Empty:
            self._resolved_layered_parameters = {}
            parameters: ParametersMap = {}
            for layer in self.ownership_layers:
                parameters.update(getattr(layer, "parameters", None) or {})

            for key, parameter in parameters.items():
                is_required = parameter.extra[EXTRA_KEY_REQUIRED]
                value_type = parameter.extra[EXTRA_KEY_VALUE_TYPE]
                if value_type is Undefined:
                    value_type = Any
                default_value = parameter.default if parameter.default is not Undefined else ...
                self._resolved_layered_parameters[key] = ModelField(
                    name=key,
                    type_=value_type,
                    field_info=parameter,
                    default=default_value,
                    model_config=ParameterConfig,
                    class_validators=None,
                    required=is_required,
                )
        return cast("Dict[str, ModelField]", self._resolved_layered_parameters)

    def resolve_guards(self) -> List[Guard]:
        """Returns all guards in the handlers scope, starting from highest to
        current layer."""
        if self._resolved_guards is Empty:
            self._resolved_guards = []
            for layer in self.ownership_layers:
                self._resolved_guards.extend(layer.guards or [])
            self._resolved_guards = cast("List[Guard]", [AsyncCallable(guard) for guard in self._resolved_guards])  # type: ignore[arg-type]
        return cast("List[Guard]", self._resolved_guards)

    def resolve_dependencies(self) -> Dict[str, Provide]:
        """Returns all dependencies correlating to handler function's kwargs
        that exist in the handler's scope."""
        if not self.signature_model:
            raise RuntimeError("resolve_dependencies cannot be called before a signature model has been generated")
        if self._resolved_dependencies is Empty:
            self._resolved_dependencies = {}
            for layer in self.ownership_layers:
                for key, value in (layer.dependencies or {}).items():
                    self._validate_dependency_is_unique(
                        dependencies=self._resolved_dependencies, key=key, provider=value
                    )
                    self._resolved_dependencies[key] = value
        return cast("Dict[str, Provide]", self._resolved_dependencies)

    def resolve_middleware(self) -> List["Middleware"]:
        """Builds the middleware stack for the RouteHandler and returns it.

        The middlewares are added from top to bottom (app -> router ->
        controller -> route handler) and then reversed.
        """
        resolved_middleware = []
        for layer in self.ownership_layers:
            resolved_middleware.extend(layer.middleware or [])
        return list(reversed(resolved_middleware))

    def resolve_exception_handlers(self) -> ExceptionHandlersMap:
        """Resolves the exception_handlers by starting from the route handler
        and moving up.

        This method is memoized so the computation occurs only once.
        """
        resolved_exception_handlers = {}
        for layer in self.ownership_layers:
            resolved_exception_handlers.update(layer.exception_handlers or {})
        return resolved_exception_handlers

    def resolve_opts(self) -> None:
        """Builds the route handler opt dictionary by going from top to bottom.

        If multiple layers define the same key, the value from the
        closest layer to the response handler will take precedence.
        """

        opt: Dict[str, Any] = {}
        for layer in self.ownership_layers:
            opt.update(layer.opt or {})

        self.opt = opt

    async def authorize_connection(self, connection: "ASGIConnection") -> None:
        """Ensures the connection is authorized by running all the route guards
        in scope."""
        for guard in self.resolve_guards():
            await guard(connection, copy(self))  # type: ignore

    @staticmethod
    def _validate_dependency_is_unique(dependencies: Dict[str, Provide], key: str, provider: Provide) -> None:
        """Validates that a given provider has not been already defined under a
        different key."""
        for dependency_key, value in dependencies.items():
            if provider == value:
                raise ImproperlyConfiguredException(
                    f"Provider for key {key} is already defined under the different key {dependency_key}. "
                    f"If you wish to override a provider, it must have the same key."
                )

    def _validate_handler_function(self) -> None:
        """Validates the route handler function once set by inspecting its
        return annotations."""
        if not self.fn:
            raise ImproperlyConfiguredException("Cannot call _validate_handler_function without first setting self.fn")

    def __str__(self) -> str:
        """
        Returns:
            A unique identifier for the route handler
        """
        target = cast("Any", self.fn)
        if not hasattr(target, "__qualname__"):
            target = type(target)
        return f"{target.__module__}.{target.__qualname__}"



from typing import Dict, Optional, Type, cast

from aiomcache import Client as MemcacheClient

from starlite.middleware.session.base import ServerSideBackend, ServerSideSessionConfig


class MemcachedBackend(ServerSideBackend["MemcachedBackendConfig"]):
    __slots__ = ("memcached",)

    def __init__(self, config: "MemcachedBackendConfig") -> None:
        """Session backend to store data in memcached.

        Notes:
            - Requires `aiomcache`. Install with `pip install starlite[memcached]`
        """
        super().__init__(config=config)
        self.memcached = config.memcached

    def _id_to_storage_key(self, session_id: str) -> bytes:
        return f"{self.config.key_prefix}:{session_id}".encode()

    async def get(self, session_id: str) -> Optional[bytes]:
        """Retrieve data associated with `session_id` from memcached.

        Args:
            session_id: The session-ID

        Returns:
            The session data, if existing, otherwise `None`.
        """
        data = await self.memcached.get(key=self._id_to_storage_key(session_id))  # type: ignore[call-overload]
        return cast("Optional[bytes]", data)

    async def set(self, session_id: str, data: bytes) -> None:
        """Store `data` in memcached under `<prefix>:<session_id>`. If there is
        already data associated with `session_id`, replace it with `data` and
        reset its expiry time.

        Args:
            session_id: The session-ID
            data: Serialized session data

        Returns:
            None
        """
        await self.memcached.set(key=self._id_to_storage_key(session_id), value=data, exptime=self.config.max_age)

    async def delete(self, session_id: str) -> None:
        """Delete the data associated with `session_id`. Fail silently if no
        such session-ID exists.

        Args:
            session_id: The session-ID

        Returns:
            None
        """
        await self.memcached.delete(self._id_to_storage_key(session_id))

    async def delete_all(self) -> None:
        """Delete all data stored within this backend.

        Returns:
            None

        Notes:
            This has poor performance since memcached does not offer utilities to
            properly scan or match keys by prefix.
        """
        stats: Dict[bytes, Optional[bytes]] = await self.memcached.stats("items")
        for key, raw_val in stats.items():
            if not raw_val:
                continue
            _, slab, field = key.split(b":")
            val = int(raw_val)
            if field != b"number" or val == 0:
                continue
            item_request: Dict[bytes, Optional[bytes]] = await self.memcached.stats(
                "cachedump", slab, str(val + 10).encode()
            )
            for keys in item_request:
                await self.memcached.delete(keys)


class MemcachedBackendConfig(ServerSideSessionConfig):
    _backend_class: Type[MemcachedBackend] = MemcachedBackend
    memcached: MemcacheClient
    """An `aiomcache.Client` instance"""
    key_prefix: str = "STARLITE_SESSION"
    """Prefix to store data under after the schema of `<prefix>:<session-ID>`"""



from contextlib import suppress
from http.cookies import _unquote as unquote_cookie
from typing import TYPE_CHECKING, Any, Dict
from urllib.parse import unquote

from orjson import JSONDecodeError, loads
from pydantic.fields import SHAPE_LIST, SHAPE_SINGLETON
from starlite_multipart.datastructures import UploadFile as MultipartUploadFile

from starlite.datastructures.upload_file import UploadFile
from starlite.enums import RequestEncodingType

if TYPE_CHECKING:

    from pydantic.fields import ModelField

    from starlite.datastructures.multi_dicts import FormMultiDict

_true_values = {"True", "true"}
_false_values = {"False", "false"}


def parse_form_data(media_type: "RequestEncodingType", form_data: "FormMultiDict", field: "ModelField") -> Any:
    """Transforms the multidict into a regular dict, try to load json on all
    non-file values.

    Supports lists.
    """
    values_dict: Dict[str, Any] = {}
    for key, value in form_data.multi_items():
        if not isinstance(value, MultipartUploadFile):
            with suppress(JSONDecodeError):
                value = loads(value)
        existing_value = values_dict.get(key)
        if isinstance(existing_value, list):
            values_dict[key].append(value)
        elif existing_value:
            values_dict[key] = [existing_value, value]
        else:
            values_dict[key] = value
    if media_type == RequestEncodingType.MULTI_PART:
        if field.shape is SHAPE_LIST:
            return list(values_dict.values())
        if field.shape is SHAPE_SINGLETON and field.type_ in (UploadFile, MultipartUploadFile) and values_dict:
            return list(values_dict.values())[0]
    return values_dict


def parse_cookie_string(cookie_string: str) -> Dict[str, str]:
    """
    Parses a cookie string into a dictionary of values.
    Args:
        cookie_string: A cookie string.

    Returns:
        A string keyed dictionary of values
    """
    output: Dict[str, str] = {}
    cookies = [cookie.split("=", 1) if "=" in cookie else ("", cookie) for cookie in cookie_string.split(";")]
    for k, v in filter(lambda x: x[0] or x[1], ((k.strip(), v.strip()) for k, v in cookies)):
        output[k] = unquote(unquote_cookie(v))
    return output



import binascii
import contextlib
import re
import time
from base64 import b64decode, b64encode
from os import urandom
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type

from orjson import dumps, loads
from pydantic import SecretBytes, validator

from starlite.datastructures import MutableScopeHeaders
from starlite.datastructures.cookie import Cookie
from starlite.exceptions import MissingDependencyException
from starlite.types import Empty

from .base import BaseBackendConfig, BaseSessionBackend

try:
    from cryptography.exceptions import InvalidTag
    from cryptography.hazmat.primitives.ciphers.aead import AESGCM
except ImportError as e:
    raise MissingDependencyException("cryptography is not installed") from e

if TYPE_CHECKING:
    from starlite.connection import ASGIConnection
    from starlite.types import Message, Scope, ScopeSession

NONCE_SIZE = 12
CHUNK_SIZE = 4096 - 64
AAD = b"additional_authenticated_data="


class CookieBackend(BaseSessionBackend["CookieBackendConfig"]):
    __slots__ = ("aesgcm", "cookie_re")

    def __init__(self, config: "CookieBackendConfig") -> None:
        """Starlite CookieSessionMiddleware.

        Args:
            config: SessionCookieConfig instance.
        """
        super().__init__(config)
        self.aesgcm = AESGCM(config.secret.get_secret_value())
        self.cookie_re = re.compile(rf"{self.config.key}(?:-\d+)?")

    def dump_data(self, data: Any, scope: Optional["Scope"] = None) -> List[bytes]:
        """Given orjson serializable data, including pydantic models and numpy
        types, dump it into a bytes string, encrypt, encode and split it into
        chunks of the desirable size.

        Args:
            data: Data to serialize, encrypt, encode and chunk.
            scope: The ASGI connection scope.

        Notes:
            - The returned list is composed of a chunks of a single base64 encoded
            string that is encrypted using AES-CGM.

        Returns:
            List of encoded bytes string of a maximum length equal to the 'CHUNK_SIZE' constant.
        """
        serialized = self.serlialize_data(data, scope)
        associated_data = dumps({"expires_at": round(time.time()) + self.config.max_age})
        nonce = urandom(NONCE_SIZE)
        encrypted = self.aesgcm.encrypt(nonce, serialized, associated_data=associated_data)
        encoded = b64encode(nonce + encrypted + AAD + associated_data)
        return [encoded[i : i + CHUNK_SIZE] for i in range(0, len(encoded), CHUNK_SIZE)]

    def load_data(self, data: List[bytes]) -> Dict[str, Any]:
        """Given a list of strings, decodes them into the session object.

        Args:
            data: A list of strings derived from the request's session cookie(s).

        Returns:
            A deserialized session value.
        """
        decoded = b64decode(b"".join(data))
        nonce = decoded[:NONCE_SIZE]
        aad_starts_from = decoded.find(AAD)
        associated_data = decoded[aad_starts_from:].replace(AAD, b"") if aad_starts_from != -1 else None
        if associated_data and loads(associated_data)["expires_at"] > round(time.time()):
            encrypted_session = decoded[NONCE_SIZE:aad_starts_from]
            decrypted = self.aesgcm.decrypt(nonce, encrypted_session, associated_data=associated_data)
            return self.deserialize_data(decrypted)
        return {}

    def get_cookie_keys(self, connection: "ASGIConnection") -> List[str]:
        """Return a list of cookie-keys from the connection if they match the
        session-cookie pattern.

        Args:
            connection: An ASGIConnection instance

        Returns:
            A list of session-cookie keys
        """
        return sorted(key for key in connection.cookies if self.cookie_re.fullmatch(key))

    def _create_session_cookies(
        self, data: List[bytes], cookie_params: Optional[Dict[str, Any]] = None
    ) -> List[Cookie]:
        """Create a list of cookies containing the session data."""
        if cookie_params is None:
            cookie_params = self.config.dict(exclude_none=True, exclude={"secret", "key"})
        return [
            Cookie(value=datum.decode("utf-8"), key=f"{self.config.key}-{i}", **cookie_params)
            for i, datum in enumerate(data)
        ]

    async def store_in_message(
        self, scope_session: "ScopeSession", message: "Message", connection: "ASGIConnection"
    ) -> None:
        """Store data from `scope_session` in `Message` in the form of cookies.
        If the contents of `scope_session` are too large to fit a single
        cookie, it will be split across several cookies, following the naming
        scheme of `<cookie key>-<n>`. If the session is empty or shrinks,
        cookies will be cleared by setting their value to `null`

        Args:
            scope_session: Current session to store
            message: Outgoing send-message
            connection: Originating ASGIConnection containing the scope

        Returns:
            None
        """

        scope = connection.scope
        headers = MutableScopeHeaders.from_message(message)
        cookie_keys = self.get_cookie_keys(connection)

        if scope_session and scope_session is not Empty:
            data = self.dump_data(scope_session, scope=scope)
            cookie_params = self.config.dict(exclude_none=True, exclude={"secret", "key"})
            for cookie in self._create_session_cookies(data, cookie_params):
                headers.add("Set-Cookie", cookie.to_header(header=""))
            # Cookies with the same key overwrite the earlier cookie with that key. To expire earlier session
            # cookies, first check how many session cookies will not be overwritten in this upcoming response.
            # If leftover cookies are greater than or equal to 1, that means older session cookies have to be
            # expired and their names are in cookie_keys.
            cookies_to_clear = cookie_keys[len(data) :] if len(cookie_keys) - len(data) > 0 else []
        else:
            cookies_to_clear = cookie_keys

        for cookie_key in cookies_to_clear:
            cookie_params = self.config.dict(exclude_none=True, exclude={"secret", "max_age", "key"})
            headers.add(
                "Set-Cookie",
                Cookie(value="null", key=cookie_key, expires=0, **cookie_params).to_header(header=""),
            )

    async def load_from_connection(self, connection: "ASGIConnection") -> Dict[str, Any]:
        """Load session data from a connection's session-cookies and return it
        as a dictionary.

        Args:
            connection: Originating ASGIConnection

        Returns:
            The session data
        """
        cookie_keys = self.get_cookie_keys(connection)
        if cookie_keys:
            data = [connection.cookies[key].encode("utf-8") for key in cookie_keys]
            # If these exceptions occur, the session must remain empty so do nothing.
            with contextlib.suppress(InvalidTag, binascii.Error):
                return self.load_data(data)
        return {}


class CookieBackendConfig(BaseBackendConfig):
    """Configuration for [SessionMiddleware] middleware."""

    _backend_class: Type[CookieBackend] = CookieBackend

    secret: SecretBytes
    """
    A secret key to use for generating an encryption key.
    Must have a length of 16 (128 bits), 24 (192 bits) or 32 (256 bits) characters.
    """

    @validator("secret", always=True)
    def validate_secret(cls, value: SecretBytes) -> SecretBytes:  # pylint: disable=no-self-argument
        """Ensures that the 'secret' value is 128, 192 or 256 bits.

        Args:
            value: A bytes string.

        Raises:
            ValueError: if the bytes string is of incorrect length.

        Returns:
            A bytes string.
        """
        if len(value.get_secret_value()) not in {16, 24, 32}:
            raise ValueError("secret length must be 16 (128 bit), 24 (192 bit) or 32 (256 bit)")
        return value



from typing import Any, Dict, Iterable, List, Mapping, Optional, Tuple, Union
from urllib.parse import parse_qsl

from multidict import MultiDict, MultiDictProxy

from starlite.datastructures.upload_file import UploadFile
from starlite.utils import deprecated


class FormMultiDict(MultiDictProxy[Any]):
    def __init__(
        self, args: Optional[Union["FormMultiDict", Mapping[str, Any], Iterable[Tuple[str, Any]]]] = None
    ) -> None:
        super().__init__(MultiDict(args or {}))

    def multi_items(self) -> List[Tuple[str, Any]]:
        """Get all keys and values, including duplicates.

        Returns:
            A list of tuples containing key-value pairs
        """
        return [(key, value) for key in set(self) for value in self.getall(key)]

    @deprecated("1.36.0", alternative="FormMultiDict.getall")
    def getlist(self, key: str) -> List[str]:
        """Get all values.

        Args:
            key: The key

        Returns:
            A list of values
        """
        return super().getall(key, [])

    async def close(self) -> None:
        """Closes all files in the multi-dict.

        Returns:
            None
        """
        for _, value in self.multi_items():
            if isinstance(value, UploadFile):
                await value.close()


class QueryMultiDict(MultiDict[Union[str, bool]]):
    def __init__(
        self, args: Optional[Union["QueryMultiDict", Mapping[str, Any], Iterable[Tuple[str, Any]]]] = None
    ) -> None:
        super().__init__(MultiDict(args or {}))

    @classmethod
    def from_query_string(cls, query_string: str) -> "QueryMultiDict":
        """Creates a QueryMultiDict from a query string.

        Args:
            query_string: A query string.

        Returns:
            A QueryMultiDict instance
        """
        _bools = {"true": True, "false": False, "True": True, "False": False}
        return cls(
            (k, v) if v not in _bools else (k, _bools[v]) for k, v in parse_qsl(query_string, keep_blank_values=True)
        )

    def dict(self) -> Dict[str, List[Any]]:
        """

        Returns:
            A dict of lists
        """
        return {k: self.getall(k) for k in set(self.keys())}



from collections import defaultdict
from traceback import format_exc
from typing import TYPE_CHECKING, Dict, List, Set, Union

from starlite.asgi.routing_trie import validate_node
from starlite.asgi.routing_trie.mapping import add_map_route_to_trie
from starlite.asgi.routing_trie.traversal import parse_scope_to_route
from starlite.asgi.routing_trie.types import create_node
from starlite.asgi.utils import get_route_handlers
from starlite.exceptions import ImproperlyConfiguredException
from starlite.utils import AsyncCallable

if TYPE_CHECKING:
    from starlite.app import Starlite
    from starlite.asgi.routing_trie.types import RouteTrieNode
    from starlite.routes import ASGIRoute, HTTPRoute, WebSocketRoute
    from starlite.routes.base import BaseRoute
    from starlite.types import (
        LifeSpanHandler,
        LifeSpanReceive,
        LifeSpanSend,
        LifeSpanShutdownCompleteEvent,
        LifeSpanShutdownFailedEvent,
        LifeSpanStartupCompleteEvent,
        LifeSpanStartupFailedEvent,
        Receive,
        RouteHandlerType,
        Scope,
        Send,
    )


class ASGIRouter:
    __slots__ = (
        "_plain_routes",
        "_registered_routes",
        "app",
        "root_route_map_node",
        "route_handler_index",
        "route_mapping",
    )

    def __init__(
        self,
        app: "Starlite",
    ) -> None:
        """This class is the Starlite ASGI router. It handles both the ASGI
        lifespan event and routing connection requests.

        Args:
            app: The Starlite app instance
        """
        self._plain_routes: Set[str] = set()
        self._registered_routes: Set[Union["HTTPRoute", "WebSocketRoute", "ASGIRoute"]] = set()
        self.app = app
        self.root_route_map_node: "RouteTrieNode" = create_node()
        self.route_handler_index: Dict[str, "RouteHandlerType"] = {}
        self.route_mapping: Dict[str, List["BaseRoute"]] = defaultdict(list)

    async def __call__(self, scope: "Scope", receive: "Receive", send: "Send") -> None:
        """The main entry point to the Router class."""
        asgi_app, handler = parse_scope_to_route(
            root_node=self.root_route_map_node, scope=scope, plain_routes=self._plain_routes
        )
        scope["route_handler"] = handler
        await asgi_app(scope, receive, send)

    def _store_handler_to_route_mapping(self, route: "BaseRoute") -> None:
        """Stores the mapping of route handlers to routes and to route handler
        names.

        Args:
            route: A Route instance.

        Returns:
            None
        """

        for handler in get_route_handlers(route):
            if handler.name in self.route_handler_index and str(self.route_handler_index[handler.name]) != str(handler):
                raise ImproperlyConfiguredException(
                    f"route handler names must be unique - {handler.name} is not unique."
                )
            identifier = handler.name or str(handler)
            self.route_mapping[identifier].append(route)
            self.route_handler_index[identifier] = handler

    async def _call_lifespan_handler(self, handler: "LifeSpanHandler") -> None:
        """Determines whether the lifecycle handler expects an argument, and if
        so passes the `app.state` to it. If the handler is an async function,
        it awaits the return.

        Args:
            handler (LifeSpanHandler): sync or async callable that may or may not have an argument.
        """
        async_callable = AsyncCallable(handler)  # type: ignore

        if async_callable.num_expected_args > 0:
            await async_callable(self.app.state)  # type: ignore[arg-type]
        else:
            await async_callable()

    def construct_routing_trie(self) -> None:
        """Create a map of the app's routes.

        This map is used in the asgi router to route requests.
        """
        new_routes = [route for route in self.app.routes if route not in self._registered_routes]
        for route in new_routes:
            node = add_map_route_to_trie(
                root_node=self.root_route_map_node,
                route=route,
                app=self.app,
                plain_routes=self._plain_routes,
            )

            if node["path_parameters"] != route.path_parameters:
                raise ImproperlyConfiguredException("Should not use routes with conflicting path parameters")

            self._store_handler_to_route_mapping(route)
            self._registered_routes.add(route)

        validate_node(node=self.root_route_map_node)

    async def lifespan(self, receive: "LifeSpanReceive", send: "LifeSpanSend") -> None:
        """Handles the ASGI "lifespan" event on application startup and
        shutdown.

        Args:
            receive: The ASGI receive function.
            send: The ASGI send function.

        Returns:
            None.
        """
        message = await receive()
        try:
            shutdown_event: "LifeSpanShutdownCompleteEvent" = {"type": "lifespan.shutdown.complete"}

            if message["type"] == "lifespan.startup":
                await self.startup()
                startup_event: "LifeSpanStartupCompleteEvent" = {"type": "lifespan.startup.complete"}
                await send(startup_event)
                await receive()
            else:
                await self.shutdown()
                await send(shutdown_event)
        except BaseException as e:
            if message["type"] == "lifespan.startup":
                startup_failure_event: "LifeSpanStartupFailedEvent" = {
                    "type": "lifespan.startup.failed",
                    "message": format_exc(),
                }
                await send(startup_failure_event)
            else:
                shutdown_failure_event: "LifeSpanShutdownFailedEvent" = {
                    "type": "lifespan.shutdown.failed",
                    "message": format_exc(),
                }
                await send(shutdown_failure_event)
            raise e
        else:
            await self.shutdown()
            await send(shutdown_event)

    async def startup(self) -> None:
        """Run any [LifeSpanHandlers][starlite.types.LifeSpanHandler] defined
        in the application's `.on_startup` list.

        Calls the `before_startup` hook and `after_startup` hook
        handlers respectively before and after calling in the lifespan
        handlers.
        """
        for hook in self.app.before_startup:
            await hook(self.app)

        for handler in self.app.on_startup:
            await self._call_lifespan_handler(handler)

        for hook in self.app.after_startup:
            await hook(self.app)

    async def shutdown(self) -> None:
        """Run any [LifeSpanHandlers][starlite.types.LifeSpanHandler] defined
        in the application's `.on_shutdown` list.

        Calls the `before_shutdown` hook and `after_shutdown` hook
        handlers respectively before and after calling in the lifespan
        handlers.
        """

        for hook in self.app.before_shutdown:
            await hook(self.app)

        for handler in self.app.on_shutdown:
            await self._call_lifespan_handler(handler)

        for hook in self.app.after_shutdown:
            await hook(self.app)



from typing import TYPE_CHECKING, Dict, List, NamedTuple, Set, Type, Union

from typing_extensions import TypedDict

if TYPE_CHECKING:
    from typing_extensions import Literal

    from starlite.types import ASGIApp, Method, RouteHandlerType
    from starlite.types.internal_types import PathParameterDefinition


class PathParameterSentinel:
    """Sentinel class designating a path parameter."""


class ASGIHandlerTuple(NamedTuple):
    """This class encapsulates a route handler node."""

    asgi_app: "ASGIApp"
    """An ASGI stack, composed of a handler function and layers of middleware that wrap it."""
    handler: "RouteHandlerType"
    """The route handler instance."""


class RouteTrieNode(TypedDict):
    """This class represents a radix trie node."""

    asgi_handlers: Dict[Union["Method", "Literal['websocket', 'asgi']"], "ASGIHandlerTuple"]
    """
    A mapping of ASGI handlers stored on the node.
    """
    child_keys: Set[Union[str, Type[PathParameterSentinel]]]
    """
    A set containing the child keys, same as the children dictionary - but as a set, which offers faster lookup.
    """
    children: Dict[Union[str, Type[PathParameterSentinel]], "RouteTrieNode"]  # type: ignore[misc]
    """
    A dictionary mapping path components or using the PathParameterSentinel class to child nodes.
    """
    is_asgi: bool
    """
    Designate the node as having an `@asgi` type handler.
    """
    is_mount: bool
    """
    Designates the node as a "mount" path, meaning that the handler function will be forwarded all sub paths.
    """
    is_path_type: bool
    """
    Designates the node as expecting a path parameter of type 'Path',
    which means that any sub path under the node is considered to be a path parameter value rather than a url.
    """
    is_static: bool
    """
    Designates the node as a static path node, which means that any sub path under the node is considered to be
    a file path in one of the static directories.
    """
    path_parameters: List["PathParameterDefinition"]
    """
    A list of tuples containing path parameter definitions. This is used for parsing extracted path parameter values.
    """


def create_node() -> "RouteTrieNode":
    """Creates a RouteMapNode instance.

    Returns:
        A route map node instance.
    """

    return {
        "asgi_handlers": {},
        "child_keys": set(),
        "children": {},
        "is_asgi": False,
        "is_mount": False,
        "is_static": False,
        "is_path_type": False,
        "path_parameters": [],
    }
